//Ethan Alexander Shulman 2022 "Ruler of Ashes" March 5, 2022

//scene rendering options
#define VIEW_STEPS 130
//#define VIEW_STEP_DITHER .3
#define VIEW_EPSILON 1e-2
//#define VIEW_EPSILON_MUL 1.2
#define LIGHTING_STEPS 100
//#define LIGHTING_STEP_DITHER .3
#define LIGHTING_EPSILON 1e-2
//#define COLOR_MAX 100.
#define DENOISE_RANGE 5
#define DENOISE_NORMAL_THRESHOLD .9
#define DENOISE_DEPTH_THRESHOLD 1e1
#define DENOISE_HASH_DIVIDE 1
#define ANIMATION_TIME 10.
#define ANIMATION_SAMPLES 1000
#ifdef ANIMATION_TIME
out float time;
#endif

//constants
#define PI 3.14159265





mat2 r2d(float a) {
	float c = cos(a), s = sin(a);
	return mat2(c,s,-s,c);
}

uvec4 hash(uvec4 seed) {
    uvec4 h = (0x6A7F8FAAu^seed)*0x01000193u;
    h = ((h.wxyz>>3u)^h^seed.yzwx)*0x01000193u;
    h = ((h.zwxy>>8u)^h^seed.wxyz)*0x01000193u;
    return h^(h>>11u);
}
#define I2F (1./float(0xFFFFFFFFu))



/*returns camera data as mat3 in the following structure {
vec3 position;
float roll, pitch, yaw;
float fieldOfViewTangent, viewNear, orbitDistance;*/
mat3 camera() {
	return mat3(70.4,15,0,
			0,0,0,
			.6,0.,0.);
}


//distance dithering for volumetrics
#define START_VOLUME(posSeed) vec2 opacityRand = fract(baseRand.zw+128.*dot(abs(posSeed),baseRand.zxy)*baseRand.xy)
#define VOLUME(dst,skip,alpha,anisotropy) max(dst,-1e3+(skip+1e3)*step(alpha+nsampling*anisotropy,opacityRand.x))

//dither materials for chromatic separation
#define START_CHROMA(posSeed) float chromaFrame = mod(baseRand.x+156.*dot(abs(posSeed),baseRand.zwy),3.)
#define CHROMA(mt,rc,gc,bc) if (chromaFrame < 1.) {mt[0]*=vec3(3,0,0);rc;}\
else if (chromaFrame < 2.) {mt[0]*=vec3(0,3,0);gc;}\
else {mt[0]*=vec3(0,0,3);bc;}



//functions defining scene
float capsule(vec3 p, vec3 a, vec3 b, float r) {
	vec3 pa = p-a, ba = b-a;
	return length(pa-ba*clamp(dot(pa,ba)/dot(ba,ba),0.,1.))-r;
}

//returns vec2(distance, materialId) of nearest surface
vec2 geometry(vec3 p) {
	//terrain
	float td = p.y;
	if (td < 100.) {
		for (float f = 1.; f < 128.; f += f*2.) {
			td -= length(mod(abs(p.xz),f)-f*.5)*.2;
		}
		if (nsampling == 1.) {
			td += //cos(dot(p.xz,vec2(46,23)))*0.+
				cos(cos(dot(p.xz,vec2(237,-12)))*12.+
				cos(dot(p.xz,vec2(127,734)))*7.)*.1;
		}
	}
	vec2 geo = vec2(td,0);
	
	//clouds
	if (td > 50.) {
		vec3 cp = mod(abs(p),200.)-100.;
		START_VOLUME(p);
		float cd = length(max(abs(cp)-54.,0));
		for (float f = 1.; f < 100.; f += f*2.) {
			cd += cos(time*PI*2./10.+dot(p,fract(f*vec3(.2734,.37624,.48734)))*20./f)*f*.1;
		}
		cd = max(abs(p.y-300.)-200.,cd);
		cd = VOLUME(cd, 1., .5, 0.);
		if (cd < geo.x) geo = vec2(cd,1);
	}
	
	//planes
	if (p.y > 50. && p.y < 150.) {
		p.z += time*30./10.;
		p.xz = mod(abs(p.xz),vec2(100.,30.))-vec2(50.,15.);
		p.y -= 100.;
		//p.x -= 60.;
		//p.z -= 20.;
		//p.y -= 20.;
		float pd = min(min(length(max(abs(p-vec3(0,1,7))-vec3(.1,1,1),0)), length(max(abs(p)-vec3(5,.1,1.3),0))),
			capsule(p,vec3(0,0,-5),vec3(0,0,7),1));
		if (pd < geo.x) geo = vec2(pd,2);
	}
	
	return geo;
}

/*returns material as mat3 in the following structure {
vec3 color;
float roughness, emission, opacity;
float refractionIndex;*/
mat3 material(vec3 p, vec3 n, float mat) {
	if (mat == 2.) {
		//plane
		return mat3(vec3(.13,.2,.2), .04,0,1, 0,0,0);
	}

	if (mat == 1.) {
		//cloud
		return mat3(vec3(.5), .4,0,1, 0,0,0);
	}

	//terrain
	return mat3(vec3(.1), .1,0,1, 0,0,0);
}

//returns background color at point and direction
vec3 background(vec3 p, vec3 d) {
	float ex = 1./max(1e-1,abs(d.x*(1.-d.y))*10.-1.);
	return mix(vec3(1,.06,.002),vec3(1,.6,.1),min(1.,ex*.1))*ex*clamp(d.y*20.,0.,1.);//mix(vec3(0), (cos(vec3(.98,.9,.84)*ex*40.)*.5+.5)*ex, clamp(d.y*10.-.5,0.,1.));
}
vec3 backgroundLighting(vec3 p, vec3 d) {
	return background(p,d);
}

